/*
* Copyright (c) 2000-2005, University of Salford
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions are met:
*
* Redistributions of source code must retain the above copyright notice, this 
* list of conditions and the following disclaimer.
* 
* Redistributions in binary form must reproduce the above copyright notice, 
* this list of conditions and the following disclaimer in the documentation 
* and/or other materials provided with the distribution. 
*
* Neither the name of the University of Salford nor the names of its 
* contributors may be used to endorse or promote products derived from this 
* software without specific prior written permission. 
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*/
/*
* Copyright (c) 2006, University of Kent
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions are met:
*
* Redistributions of source code must retain the above copyright notice, this 
* list of conditions and the following disclaimer.
* 
* Redistributions in binary form must reproduce the above copyright notice, 
* this list of conditions and the following disclaimer in the documentation 
* and/or other materials provided with the distribution. 
*
* 1. Neither the name of the University of Kent nor the names of its 
* contributors may be used to endorse or promote products derived from this 
* software without specific prior written permission. 
*
* 2. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS  
* IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
* THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
* PURPOSE ARE DISCLAIMED. 
*
* 3. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
* 4. YOU AGREE THAT THE EXCLUSIONS IN PARAGRAPHS 2 AND 3 ABOVE ARE REASONABLE
* IN THE CIRCUMSTANCES.  IN PARTICULAR, YOU ACKNOWLEDGE (1) THAT THIS
* SOFTWARE HAS BEEN MADE AVAILABLE TO YOU FREE OF CHARGE, (2) THAT THIS
* SOFTWARE IS NOT "PRODUCT" QUALITY, BUT HAS BEEN PRODUCED BY A RESEARCH
* GROUP WHO DESIRE TO MAKE THIS SOFTWARE FREELY AVAILABLE TO PEOPLE WHO WISH
* TO USE IT, AND (3) THAT BECAUSE THIS SOFTWARE IS NOT OF "PRODUCT" QUALITY
* IT IS INEVITABLE THAT THERE WILL BE BUGS AND ERRORS, AND POSSIBLY MORE
* SERIOUS FAULTS, IN THIS SOFTWARE.
*
* 5. This license is governed, except to the extent that local laws
* necessarily apply, by the laws of England and Wales.
*/

package issrg.aef;

import issrg.pba.*;
import issrg.pba.rbac.*;
import issrg.utils.RFC2253NameParser;
import javax.naming.Context;
import javax.naming.directory.InitialDirContext;

import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.io.ByteArrayInputStream;

/**
* This is a sample code for the AEF.
*/

public class SampleAEF1 {
    static { // add ShibbolethURLHandler only once when the JVM starts
      try{ 
        CustomisePERMIS.addURLHandler("issrg.shibboleth.ShibbolethURLHandler");
      }catch (ClassNotFoundException cnfe){
	System.out.println("Shibboleth URL handler was not found - Shibboleth policies will not be handled");
      }
      CustomisePERMIS.registerACExtensions();
    }

	private final static java.io.BufferedReader in = new java.io.BufferedReader(new java.io.InputStreamReader(System.in));
	private static final String PKC="rootca.cer";
	private static final String PERMIS_EPF="permis.epf";
	private static final String PERMIS_PSW="l3tM3InNow";
	private static final String SALFORD_LDAP="ldap://sec.cs.kent.ac.uk/";
	private static final String AC_ATTRIBUTE="attributeCertificateAttribute";
	private static final String PKC_ATTRIBUTE="userCertificate";

	private static String SOA = "";
	private static String oID = "";
	private static String LDAP = "";
	private static String uRL = "";
	private static String AC_attribute = "";
	private static String PKC_attribute = "";
	private final static issrg.utils.repository.VirtualRepository vr = new issrg.utils.repository.VirtualRepository();

	private static CertificateFactory cf;

	public static void main(String [] args){
		try{
			if (args.length<1){
				SOA = readLine("SOA DN: ");
				oID = readLine("Policy OID: ");
				LDAP = readLine("LDAP URL (leave empty to use Kent University LDAP): ");
				if (LDAP.intern()!=""){
					AC_attribute = readLine("Attribute Certificate attribute name (will use "+AC_ATTRIBUTE+" if left empty): ");
					PKC_attribute = readLine("User Certificate attribute name (will use "+PKC_ATTRIBUTE+" if left empty): ");
				}

				while (true){
					String filename = readLine("AC file (empty line to stop loop): ");
					if (filename.intern() == "") break;

					loadAC(vr, filename);
				}

				while (true){
					String filename = readLine("PKC file (empty line to stop loop): ");
					if (filename.intern()=="") break;

					loadPKC(vr, filename);
				}
    			}else{
				java.io.BufferedReader in=null;


				try{
					if (args.length>1) System.setErr(new java.io.PrintStream(new java.io.FileOutputStream(args[1])));
				}catch(java.io.IOException io){
					System.err.println("Could not open file "+args[1]);
					throw io;
				}

				try{
					in = new java.io.BufferedReader(new java.io.FileReader(args[0]));
				}catch(java.io.IOException io){
					System.err.println("Could not open file "+args[0]);
					throw io;
				}


				String s;

				while ((s=in.readLine())!=null){
					int i = s.indexOf('=');	// find the assignment mark
					String varName = null;

					if (i>=0){
						varName = s.substring(0, i).intern();
						s = s.substring(i+1);
					}

					if (varName == "SOA"){
						SOA = s;
					}else
					if (varName == "OID"){
						oID = s;
					}else
					if (varName == "LDAP_AC_attribute"){
						AC_attribute = s;
					}else
					if (varName == "LDAP_PKC_attribute"){
						PKC_attribute = s;
					}else
					if (varName == "LDAP"){
						LDAP = s;
					}else
					if (varName == "URL"){
						uRL = s;
					}else
					if (varName == "PKC"){
						loadPKC(vr, s);
					}else
     					if (varName == "AC"){
						loadAC(vr, s);
					}else
						println("Unrecognised line; ignored: "+varName+"="+s);
				}

			}

			if (LDAP.intern()==""){
				LDAP=SALFORD_LDAP;
			}
			if (AC_attribute.intern()==""){
				AC_attribute=AC_ATTRIBUTE;
			}
			if (PKC_attribute.intern()==""){
				PKC_attribute=PKC_ATTRIBUTE;
			}

			issrg.utils.repository.AttributeRepository r=vr;

			String ch = readLine("Choose repository (type a letter or press return for default):\nl - to use LDAP repository ("+LDAP+")\nu - specify a repository URL ("+(uRL.intern()==""?"you will be prompted to type the URL":uRL)+")\n(default) - press enter to use ACs from files\n?: ");
			if (ch.intern()=="l"){
				javax.naming.directory.DirContext [] ldaps = new javax.naming.directory.DirContext[1];
				java.util.Hashtable env = new java.util.Hashtable();
				env.put(Context.INITIAL_CONTEXT_FACTORY, "com.sun.jndi.ldap.LdapCtxFactory");
				env.put(Context.PROVIDER_URL, LDAP);
      						// put your own URL there
				env.put("java.naming.ldap.version", "3");
      		// change the version, if you are not running v3 LDAP server
				env.put("java.naming.ldap.attributes.binary", (issrg.pba.repository.ACRepository.ATTRIBUTE_CERTIFICATE_ID=AC_attribute)+" "+(issrg.security.PKCRepository.USER_PKC_ATTRIBUTE=PKC_attribute));
     		// otherwise it may sometimes retrieve it as a String

				ldaps[0] = new InitialDirContext(env);

				r=new issrg.utils.repository.LDAPRepository(ldaps);
			}else if (ch.intern()=="u"){
				if (uRL.intern()==""){
					uRL = readLine("Type the Repository URL (file: or ldap: are supported): ");
				}

				r=issrg.pba.rbac.URLHandler.getRepositoryByURL(uRL);
				if (r==null) throw new BadURLException("No URL Handler was found for this protocol: "+uRL);
			}


			// here the SOA and the OID are known

			issrg.pba.rbac.SignatureVerifier sv = new SamplePKI();

			String epf = readLine("Choose cryptography (type a letter or press return for default):\ne - Entrust\ns - Simple (X.509 PKC) signature verification\n(default) - no authentication\n?: ");
			if (epf.intern()!=""){
			  if (epf.intern()=="e"){
			    epf = readLine("Enter location of the epf file\n[press return to use the default permis.epf]: ").intern();
			    sv=new issrg.pba.rbac.SimpleSignatureVerifier(new issrg.security.EntrustSecurity(epf==""?PERMIS_EPF:epf, new com.entrust.toolkit.util.SecureStringBuffer(epf==""?PERMIS_PSW:readLine("Input password: "))));
  			  }else{
			    if (cf==null) cf=CertificateFactory.getInstance("X.509");
			    epf = readLine("Enter location of the PKC file of Root CA ["+PKC+"]: ").intern();
			    X509Certificate x509 = (X509Certificate)cf.generateCertificate(new java.io.FileInputStream(epf==""?PKC:epf));
			    issrg.security.DefaultVerifier ds = new issrg.security.DefaultVerifier();
			    ds.setRootCA(x509);
			    ds.setPKCRepository(new issrg.security.PKCRepository(r));
			    sv=new issrg.pba.rbac.SimpleSignatureVerifier(ds);
			  }
			}
			PBAAPI pbaApi = new PermisRBAC(new iaik.asn1.ObjectID(oID), new LDAPDNPrincipal(SOA),
					r, sv);
			SampleAEF1 aef = new SampleAEF1(pbaApi);

			String userDN = "";
			String targetDN = "";
			String action = "";

			while(true){
				String s = readLine("\n\n*** Action Request ***\nUser DN or \".\" to quit ["+userDN+"]: ").intern();
				if (s == ".") break;

				if (s != ""){
					userDN = s;
				}

				s = readLine("Target DN or its URL ["+targetDN+"]: ").intern();

				if (s != ""){
					targetDN = s;
				}

				s = readLine("Action ["+action+"]: ").intern();

				if (s != ""){
					action = s;
				}

				PermisTarget pt = null;
				try{
					pt=new PermisTarget(targetDN); // see if it is a URL or a DN
				}catch (issrg.pba.rbac.BadURLException bue){
					pt=new PermisTarget(targetDN, null);
				}

				println(aef.execute(new LDAPDNPrincipal(userDN),
						new PermisAction(action),//, new issrg.pba.rbac.Argument[]{new issrg.pba.rbac.PermisArgument("Integer", "1")}),
						pt)
					);
			}
		}catch (Throwable th){
			println("Failed to initialise: "+th.getMessage());
			th.printStackTrace();
		}
		System.exit(0);
	}

	public static void loadAC(issrg.utils.repository.VirtualRepository vr, String filename){
		try{
			java.io.InputStream io = new java.io.FileInputStream(filename);
			byte [] ac = new byte [io.available()];

			io.read(ac);

			issrg.ac.AttributeCertificate acd = issrg.ac.AttributeCertificate.guessEncoding(ac);

			// now put the ac into the entry with the name of the holder
			// we are working with explicit Holder specification only: no BCIDs

			vr.populate(issrg.ac.Util.generalNamesToString(acd.getACInfo().getHolder().getEntityName()).toUpperCase(),
				issrg.pba.repository.ACRepository.ATTRIBUTE_CERTIFICATE_ID,
				ac);
		}catch (Throwable th){
			println("Failed to load AC from ["+filename+"]");
			th.printStackTrace();
		}
	}

	public static void loadPKC(issrg.utils.repository.VirtualRepository vr, String filename){
		try{
			if (cf==null) cf=CertificateFactory.getInstance("X.509");

			java.io.InputStream io = new java.io.FileInputStream(filename);
			byte [] pkc = new byte [io.available()];

			io.read(pkc);

			X509Certificate x509 = (X509Certificate)cf.generateCertificate(new ByteArrayInputStream(pkc));
			//System.out.println("loading PKC for "+x509.getSubjectDN().getName());//******************

			String s=RFC2253NameParser.toCanonicalDN(RFC2253NameParser.distinguishedName(x509.getSubjectDN().getName()));

			vr.populate(s, issrg.security.PKCRepository.USER_PKC_ATTRIBUTE, pkc);
		}catch (Throwable th){
			println("Failed to load PKC from ["+filename+"]");
			th.printStackTrace();
		}
	}

 	public static String readLine(String prompt){
		System.out.print(prompt);
		String s = null;

		try{
			s=in.readLine();
		}catch(java.io.IOException ioe){
		}

		return s==null?"":s;
	}

	public static void println(String s){
		// add code for logging things to a file
		System.out.println(s);
	}



	PBAAPI pbaApi = null;

	public SampleAEF1(PBAAPI pbaApi){
		this.pbaApi = pbaApi;
	}

	public String execute(java.security.Principal user, Action action, Target target){
		try{
			authenticate(user);

			// no cache is known
			Subject s = pbaApi.getCreds(user, null);

			if (!pbaApi.decision(s, action, target, null)){
				return "1: the action is not allowed";
			}

			// here you call the action on the target
		}catch (PbaException pe){
			pe.printStackTrace();
			return "2: invalid input: "+pe.getMessage();
		}catch (Throwable th){
			return "3: run-time error: "+th.getMessage();
		}

		return "0: action succeeded";
	}

	public void authenticate(java.security.Principal user){
		println ("Authenticate "+user.getName()+": [auto]");
	}
}